:
# NAME:
#	shpp - Bourne shell pre-processor
#
# SYNOPSIS:
#	shpp
#
# DESCRIPTION:
#	This module provides 'shpp' a simple? shell pre-processor.
#	It provides the recursive funcion 'shpp' as well as the actual
#	implementation in '_shpp'.
#	
#	'shpp' reads lines like:
#.nf
#
#		"op" "rest"
#.fi
#
#	and passes them to '_shpp', "op" is one of:
#
#	'.if'	Evaluates "rest" and sets "shpp_skip" accordingly,
#		and increments the nesting level.
#
#	'.elif'	If "shpp_skip" is set and we are at the right nesting
#		level, evaluates "rest" and sets "shpp_skip" accordingly.
#		It does NOT modify the nesting level. ('.elseif' also
#		accepted).
#
#	'.else'	If at the correct nesting level, and not skipping to
#		the '.endif', toggles "shpp_skip".
#
#	'.endif'
#		Decrements the nesting level and resets "shpp_skip".
#		('.fi' also accepted).
#
#	'.eval'
#		Simply evals "rest".  Skipped if "DOIT" is set.
#
#	'.set'
#		Same as '.eval', but not skipped if "DOIT" is set.
#		Should only be used for commands that will not effect
#		anything - such as simply setting variables.
#
#	'.include'
#		Evals "SHPP_INCLUDE rest" | 'shpp'
#		the default value for "SHPP_INCLUDE" is 'cat'.
#
#	'.require'
#		Like '.include' except that if the argument has
#		already been '.include'ed then nothing is done.
#
#	All other records are skipped or echoed based on "shpp_skip".
#
#	If the first arg to '.if', '.elif' etc., is '!' then the sense
#	of the condition is reversed - failure is expected.
#
# BUGS:
#	Not incredibly quick :-)
#
# AUTHOR:
#	Simon J. Gerraty <sjg@crufty.net>
#

# RCSid:
#	$Id: shpp.sh,v 1.12 2023/03/18 17:28:01 sjg Exp $
#
#	@(#) Copyright (c) 1995-2023 Simon J. Gerraty
#
#	This file is provided in the hope that it will
#	be of use.  There is absolutely NO WARRANTY.
#	Permission to copy, redistribute or otherwise
#	use this file is hereby granted provided that 
#	the above copyright notice and this notice are
#	left intact. 
#      
#	Please send copies of changes and bug-fixes to:
#	sjg@crufty.net
#

Myname=${Myname:-`basename $0 .sh`}
TMP=${TMP:-/tmp}
INCLUDED_LIST=${INCLUDED_LIST:-$TMP/$Myname.included}

# if we have debug.sh we want to use it
_have_debug_sh() {
	test x$_DEBUG_SH != x
}

_shpp_if() {
	nest=${nest:-0}
	
	set -- $*
	case "$1" in
	!)	_e="!="; shift;;
	*)	_e="=";;
	esac
	eval "$@"
	if [ $? $_e 0 ]; then
		shpp_skip=
		t=.endif$nest
	else
		shpp_skip=yes
		t=.el$nest
	fi
	eval shpp_skipto$nest=$t
}

_skipto() {
	eval "skipto=\$shpp_skipto$nest1"
}

_shpp_include() {
	[ "$INCLUDED_LIST" ] && echo $* >> $INCLUDED_LIST
	${SHPP_CAT:-cat} $* | shpp
}

# do this as a separate function
_shpp() {
	op=$1; shift
	rest="$@"

	did_x=
	if _have_debug_sh; then
		DebugOn $Myname:shpp $Myname:shpp:$op $Myname:$op shpp shpp:$op
		$DEBUG_DO eval did_x=yes
	else
		case ",$DEBUG," in
		*,$Myname:.if,*|*,$Myname:shpp,*|*,shpp,*|*,$Myname:shpp:$op,*|*,shpp:$op,*)
			case $op in
			.*)
				echo "+ $op $rest" >&2
				set -x
				did_x=yes
			esac
			;;
		esac
	fi
		
	case "$op" in
	.if)
		if [ "$shpp_skip" ]; then
			eval shpp_skipto$nest=\$shpp_skipto$nest1
		else
			_shpp_if $rest
		fi
		nest1=${nest:-0}
		nest=`expr ${nest:-0} + 1`
		;;
	.elif|.elsif|.elseif)
		_skipto
		if [ "$shpp_skip" ]; then
			if [ ${skipto:-x} = .el${nest1:-0} ]; then
				_shpp_if $rest
			fi
		elif [ ${skipto:-x} = .endif${nest1:-0} ]; then
			shpp_skip=yes
		fi
		# note we do _not_ adjust nest
		;;
	.else)
		_skipto
		if [ $skipto = .el${nest1:-0} ]; then
			[ "$shpp_skip" ] && shpp_skip= || shpp_skip=yes
		elif [ ${skipto:-x} = .endif${nest1:-0} ]; then
			shpp_skip=yes
		fi    
		;;
	.endif|.fi)
		if [ "$shpp_skip" ]; then
			_skipto
			case "${skipto:-x}" in
			.e*${nest1:-0})
				shpp_skip=
				;;
			esac
		fi
		nest=${nest1:-0}
		case "$nest" in
		0) ;;
		*)	nest1=`expr ${nest:-0} - 1`;;
		esac
		;;
	.eval)
		[ "$shpp_skip" ] || {
			[ "x$DOIT" = x ] && eval "$rest" || echo "$rest"
		}
		;;
	.set)
		[ "$shpp_skip" ] || eval "$rest"
		;;
	.include)
		[ "$shpp_skip" ] || eval ${SHPP_INCLUDE:-_shpp_include} $rest
		;;
	.require)
		# like .include but only actioned if not already done
		[ "$shpp_skip" ] || {
			if [ -s ${INCLUDED_LIST:-/dev/null} ]; then
				case `grep "^$rest" $INCLUDED_LIST` in
				"") ;;
				*) shpp_skip=:;; # borrow this
				esac
			fi
			[ "$shpp_skip" ] ||
			eval ${SHPP_INCLUDE:-_shpp_include} $rest
			shpp_skip= # reset
		}
		;;
	esac
	_have_debug_sh &&
	DebugOff $Myname:shpp $Myname:shpp:$op $Myname:$op shpp shpp:$op
}
	
shpp() {
	_have_debug_sh &&
	DebugOn $Myname:shpp shpp

	while read op rest
	do
		# always skip blank lines and comments.
		case "$op" in
		"")	continue;;
		\#*)	continue;;
		.*)
			if [ "$SHPP_DO" ]; then
				#
				# this lets us do .include's now,
				# but defer .if etc until later
				#
				_do=
				# we can't do it all in the eval as we
				# don't want to eval $rest (yet)
				eval "case ',$SHPP_DO,' in *,$op,*) _do=yes;; esac"
				case "$_do" in
				yes)	;; # below
				*)	# just pass it on
					echo "$op $rest"
					continue
					;;
				esac
			fi
			;;
		*)	# a normal record, should we skip it
			if [ "$shpp_skip" ]; then
				# set this to a function if you want...
				${SHPP_SKIP:-:} "$op" "$rest"
			else
				echo "$op" "$rest"
			fi
			continue
			;;
		esac

		if [ "$SHPP_PASSON" ]; then
			${SHPP_PASSON:-:} "$op" "$rest"
			# some callers might not want everything eval'd now
			eval "case ',$SHPP_DEFER,' in *,$op,*) continue;; esac"
		fi
		_shpp "$op" "$rest"
	done
	_have_debug_sh &&
	DebugOff $Myname:shpp shpp
}

case "$0" in
*shpp|*shpp.sh) cat $* | shpp;;
esac
